/*
 * Routines and data structures to parse URLs
 *
 * Author: Bert Bos <bert@w3.org>
 * Created: 7 March 1999
 * Version: $Id: url.c,v 1.4 1999/07/31 19:45:12 bbos Exp $
 */

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <regex.h>
#include "export.h"
#include "heap.e"

EXPORT typedef struct {
  char *full;					/* Full URL as a string */
  char *proto;					/* Protocol */
  char *user;					/* User name */
  char *password;				/* User's password */
  char *machine;				/* Domain name or IP number */
  char *port;					/* Port number or service */
  char *path;					/* Path part of URL */
  char *fragment;				/* Fragment ID part of URL */
} *URL;


/* URL_dispose -- free the memory used by a URL struct */
EXPORT void URL_dispose(URL url)
{
  if (url) {
    dispose(url->full);
    dispose(url->user);
    dispose(url->password);
    dispose(url->machine);
    dispose(url->port);
    dispose(url->path);
    dispose(url->fragment);
    dispose(url);
  }
}

/* URL_new -- create a new URL struct; return NULL if not a valid URL */
EXPORT URL URL_new(const char *url)
{
  static const char* pat =
    "^(([a-z]+):)?(//(([^:@]+)(:([^@]+))?@)?([^:/]+)(:([0-9]+))?)?([^#]*)?(#(.*))?|([^#]*)(#(.*))?$";
  /*  12      2 1 3  45      56 7     76  4 8      89 A      A9 3 B      B C D  DC  E     EF G  GF
   * 2 = proto, 5 = user, 7 = passwd, 8 = machine, A = port, B = path, D = fragment
   * E = path in case it is not a well-formed URL, G = fragment if ditto
   */
# define MAXSUB 18
  static regex_t re;
  static int initialized = 0;
  regmatch_t pm[MAXSUB];
  URL result;

  assert(url != NULL);
  
  /* Compile the regexp, only once */
  if (! initialized) {
    assert(regcomp(&re, pat, REG_EXTENDED) == 0); /* Could be memory, though... */
    initialized = 1;
  }

  /* Match the URL against the pattern; return NULL if no match */
  if (regexec(&re, url, MAXSUB, pm, 0) != 0) return NULL;

  /* Store the various parts */
  new(result);
  result->full = newstring(url);
  result->proto = pm[2].rm_so == -1
    ? NULL : newnstring(url, pm[2].rm_eo);
  result->user = pm[5].rm_so == -1
    ? NULL : newnstring(url + pm[5].rm_so, pm[5].rm_eo - pm[5].rm_so);
  result->password = pm[7].rm_so == -1
    ? NULL : newnstring(url + pm[7].rm_so, pm[7].rm_eo - pm[7].rm_so);
  result->machine = pm[8].rm_so == -1
    ? NULL : newnstring(url + pm[8].rm_so, pm[8].rm_eo - pm[8].rm_so);
  result->port = pm[10].rm_so == -1
    ? NULL : newnstring(url + pm[10].rm_so, pm[10].rm_eo - pm[10].rm_so);
  result->path = pm[11].rm_so == -1
    ? NULL : newnstring(url + pm[11].rm_so, pm[11].rm_eo - pm[11].rm_so);
  result->fragment = pm[13].rm_so == -1
    ? NULL : newnstring(url + pm[13].rm_so, pm[13].rm_eo - pm[13].rm_so);
  if (pm[14].rm_so != -1) {
    result->path = newnstring(url + pm[14].rm_so, pm[14].rm_eo - pm[14].rm_so);
    result->fragment = pm[16].rm_so == -1
      ? NULL : newnstring(url + pm[16].rm_so, pm[16].rm_eo - pm[16].rm_so);
  }
  return result;
}

/* merge_path -- make path absolute */
static char * merge_path(const char *base, const char *path)
{
  static regex_t re;
  static int initialized = 0;
  regmatch_t pm[1];
  char *p, *s;
  size_t len;

  assert(*path != '/');
  if (base && (p = strrchr(base, '/'))) {
    newarray(s, (p - base) + strlen(path) + 2);
    strncpy(s, base, (p - base) + 1);
    strcpy(s + ((p - base) + 1), path);
  } else {
    newarray(s, strlen(path) + 2);
    s[0] = '/';
    strcpy(s + 1, path);
  }

  /* Replace all substrings of form "/xxx/../" with "/" */
  if (! initialized) {
    assert(regcomp(&re, "/[^/]+/\\.\\./", REG_EXTENDED) == 0);
    initialized = 1;
  }
  len = strlen(s);
  while (regexec(&re, s, 1, pm, 0) == 0) {
    memmove(s + pm[0].rm_so, s + (pm[0].rm_eo - 1), len - pm[0].rm_eo + 2);
    len -= pm[0].rm_eo - pm[0].rm_so - 1;
  }

  /* Replace all substrings of the form "/./" with "/" */
  /*len = strlen(s);*/
  for (p = s; (p = strstr(p, "/./")); ) {
    memmove(p, p + 2, len - (p - s) - 1);
    len -= 2;
  }

  return s;    
}

/* URL_absolutize -- make a relative URL absolute */
EXPORT URL URL_absolutize(const URL base, const URL url)
{
  URL abs;

  new(abs);
  if (url->proto) {				/* Not relative at all */
    abs->proto = newstring(url->proto);
    abs->user = newstring(url->user);
    abs->password = newstring(url->password);
    abs->machine = newstring(url->machine);
    abs->port = newstring(url->port);
    abs->path = newstring(url->path);
  } else {					/* Missing proto */
    abs->proto = newstring(base->proto);	/* Copy proto from base */
    if (url->machine) {				/* Missing machine */
      abs->user = newstring(url->user);
      abs->password = newstring(url->password);
      abs->machine = newstring(url->machine);
      abs->port = newstring(url->port);
      abs->path = newstring(url->path);
    } else {
      abs->user = newstring(base->user);
      abs->password = newstring(base->password);
      abs->machine = newstring(base->machine);
      abs->port = newstring(base->port);
      if (!url->path || *(url->path) == '/') {	/* Absolute path */
	abs->path = newstring(url->path);
      } else {
	abs->path = merge_path(base->path, url->path);
      }
    }
  }
  abs->fragment = newstring(url->fragment);

  newarray(abs->full, (abs->proto ? strlen(abs->proto) + 1 : 0)
	   + (abs->user ? strlen(abs->user) + 1 : 0)
	   + (abs->password ? strlen(abs->password) + 1 : 0)
	   + (abs->machine ? strlen(abs->machine) + 2 : 0)
	   + (abs->port ? strlen(abs->port) + 1 : 0)
	   + (abs->path ? strlen(abs->path) : 0)
	   + (abs->fragment ? strlen(abs->fragment) + 1 : 0)
	   + 1);
  sprintf(abs->full, "%s%s%s%s%s%s%s%s%s%s%s%s%s",
	  abs->proto ? abs->proto : "",
	  abs->proto ? ":" : "",
	  abs->machine ? "//" : "",
	  abs->user ? abs->user : "",
	  abs->password ? ":" : "",
	  abs->password ? abs->password : "",
	  abs->user ? "@" : "",
	  abs->machine ? abs->machine : "",
	  abs->port ? ":" : "",
	  abs->port ? abs->port : "",
	  abs->path ? abs->path : "",
	  abs->fragment ? "#" : "",
	  abs->fragment ? abs->fragment : "");

  return abs;
}

/* URL_s_absolutize -- make a relative URL absolute */
EXPORT char *URL_s_absolutize(const char *base, const char *url)
{
  URL url1 = URL_new(url), base1 = URL_new(base);
  URL abs = URL_absolutize(base1, url1);
  char *result = newstring(abs->full);
  URL_dispose(abs);
  URL_dispose(url1);
  URL_dispose(base1);
  return result;
}
